@@ -75,8 +75,10 @@ void list_disk_geometry(struct fdisk_context *cxt)
 		fdisk_info(cxt, _("Disklabel type: %s"),
 				fdisk_label_get_name(lb));
 
-	if (!fdisk_is_details(cxt) && fdisk_get_disklabel_id(cxt, &id) == 0 && id)
+	if (!fdisk_is_details(cxt) && fdisk_get_disklabel_id(cxt, &id) == 0 && id) {
 		fdisk_info(cxt, _("Disk identifier: %s"), id);
+		free(id);
+	}
 }
 
 void list_disklabel(struct fdisk_context *cxt)
@@ -141,7 +143,7 @@ void list_disklabel(struct fdisk_context *cxt)
 		if (!co)
 			goto done;
 
-		/* set colum header color */
+		/* set column header color */
 		if (bold)
 			scols_cell_set_color(scols_column_get_header(co), bold);
 	}
@@ -405,9 +407,7 @@ void list_available_columns(FILE *out)
 	if (!cxt)
 		return;
 
-	termwidth = get_terminal_width();
-	if (termwidth <= 0)
-		termwidth = 80;
+	termwidth = get_terminal_width(80);
 
 	fprintf(out, _("\nAvailable columns (for -o):\n"));
 
