@@ -48,14 +48,16 @@
 #include <utmp.h>
 #include <stdlib.h>
 #include <sys/syslog.h>
-#include <sys/sysmacros.h>
 #ifdef HAVE_LINUX_MAJOR_H
 # include <linux/major.h>
 #endif
 #include <netdb.h>
-#include <lastlog.h>
 #include <security/pam_appl.h>
-#include <security/pam_misc.h>
+#ifdef HAVE_SECURITY_PAM_MISC_H
+# include <security/pam_misc.h>
+#elif defined(HAVE_SECURITY_OPENPAM_H)
+# include <security/openpam.h>
+#endif
 #include <sys/sendfile.h>
 
 #ifdef HAVE_LIBAUDIT
@@ -67,9 +69,11 @@
 #include "pathnames.h"
 #include "strutils.h"
 #include "nls.h"
+#include "env.h"
 #include "xalloc.h"
 #include "all-io.h"
 #include "fileutils.h"
+#include "ttyutils.h"
 
 #include "logindefs.h"
 
@@ -109,13 +113,13 @@ struct login_context {
 	char		vcsan[VCS_PATH_MAX];
 #endif
 
-	char		thishost[MAXHOSTNAMELEN + 1];	/* this machine */
-	char		*thisdomain;			/* this machine domain */
-	char		*hostname;			/* remote machine */
-	char		hostaddress[16];		/* remote address */
+	char		*thishost;		/* this machine */
+	char		*thisdomain;		/* this machine's domain */
+	char		*hostname;		/* remote machine */
+	char		hostaddress[16];	/* remote address */
 
 	pid_t		pid;
-	int		quiet;		/* 1 is hush file exists */
+	int		quiet;		/* 1 if hush file exists */
 
 	unsigned int	remote:1,	/* login -h */
 			nohost:1,	/* login -H */
@@ -124,7 +128,7 @@ struct login_context {
 };
 
 /*
- * This bounds the time given to login.  Not a define so it can
+ * This bounds the time given to login.  Not a define, so it can
  * be patched on machines where it's too small.
  */
 static unsigned int timeout = LOGIN_TIMEOUT;
@@ -153,7 +157,7 @@ static int is_consoletty(int fd)
  * soaking up pts's.  What they seem to hung up on is trying to write out the
  * message 'Login timed out after %d seconds' when the connection has already
  * been dropped.
- * What I did was add a second timeout while trying to write the message so
+ * What I did was add a second timeout while trying to write the message, so
  * the process just exits if the second timeout expires.
  */
 static void __attribute__ ((__noreturn__))
@@ -181,11 +185,11 @@ static void timedout(int sig __attribute__ ((__unused__)))
 
 /*
  * This handler allows to inform a shell about signals to login. If you have
- * (root) permissions you can kill all login childrent by one signal to login
- * process.
+ * (root) permissions, you can kill all login children by one signal to the
+ * login process.
  *
- * Also, parent who is session leader is able (before setsid() in child) to
- * inform child when controlling tty goes away (e.g. modem hangup, SIGHUP).
+ * Also, a parent who is session leader is able (before setsid() in the child)
+ * to inform the child when the controlling tty goes away (e.g. modem hangup).
  */
 static void sig_handler(int signal)
 {
@@ -198,8 +202,8 @@ static void sig_handler(int signal)
 }
 
 /*
- * Let use delay for all exit() calls when user is not authenticated or
- * session fully initialized (loginpam_session()).
+ * Let us delay all exit() calls when the user is not authenticated
+ * or the session not fully initialized (loginpam_session()).
  */
 static void __attribute__ ((__noreturn__)) sleepexit(int eval)
 {
@@ -209,13 +213,13 @@ static void __attribute__ ((__noreturn__)) sleepexit(int eval)
 
 static const char *get_thishost(struct login_context *cxt, const char **domain)
 {
-	if (!*cxt->thishost) {
-		if (gethostname(cxt->thishost, sizeof(cxt->thishost))) {
+	if (!cxt->thishost) {
+		cxt->thishost = xgethostname();
+		if (!cxt->thishost) {
 			if (domain)
 				*domain = NULL;
 			return NULL;
 		}
-		cxt->thishost[sizeof(cxt->thishost) -1] = '\0';
 		cxt->thisdomain = strchr(cxt->thishost, '.');
 		if (cxt->thisdomain)
 			*cxt->thisdomain++ = '\0';
@@ -227,12 +231,12 @@ static const char *get_thishost(struct login_context *cxt, const char **domain)
 }
 
 /*
- * Output the /etc/motd file
+ * Output the /etc/motd file.
  *
- * motd() determines the name of a login announcement file and outputs it to
- * the user's terminal at login time.  The MOTD_FILE configuration option is a
- * colon-delimited list of filenames. The empty MOTD_FILE option disables motd
- * printing at all.
+ * It determines the name of a login announcement file and outputs it to the
+ * user's terminal at login time.  The MOTD_FILE configuration option is a
+ * colon-delimited list of filenames.  An empty MOTD_FILE option disables
+ * message-of-the-day printing completely.
  */
 static void motd(void)
 {
@@ -251,13 +255,11 @@ static void motd(void)
 		struct stat st;
 		int fd;
 
-		if (stat(motdfile, &st) || !st.st_size)
-			continue;
 		fd = open(motdfile, O_RDONLY, 0);
 		if (fd < 0)
 			continue;
-
-		sendfile(fileno(stdout), fd, NULL, st.st_size);
+		if (!fstat(fd, &st) && st.st_size)
+			sendfile(fileno(stdout), fd, NULL, st.st_size);
 		close(fd);
 	}
 
@@ -265,15 +267,15 @@ static void motd(void)
 }
 
 /*
- * Nice and simple code provided by Linus Torvalds 16-Feb-93
- * Nonblocking stuff by Maciej W. Rozycki, macro@ds2.pg.gda.pl, 1999.
+ * Nice and simple code provided by Linus Torvalds 16-Feb-93.
+ * Non-blocking stuff by Maciej W. Rozycki, macro@ds2.pg.gda.pl, 1999.
  *
  * He writes: "Login performs open() on a tty in a blocking mode.
  * In some cases it may make login wait in open() for carrier infinitely,
  * for example if the line is a simplistic case of a three-wire serial
- * connection. I believe login should open the line in the non-blocking mode
+ * connection. I believe login should open the line in non-blocking mode,
  * leaving the decision to make a connection to getty (where it actually
- * belongs).
+ * belongs)."
  */
 static void open_tty(const char *tty)
 {
@@ -337,7 +339,7 @@ static void chown_tty(struct login_context *cxt)
 		if (chmod(cxt->vcsn, cxt->tty_mode))
 			chmod_err(cxt->vcsn, cxt->tty_mode);
 
-		if (chown(cxt->vcsan, uid, gid))			/* vcsa */
+		if (chown(cxt->vcsan, uid, gid))		/* vcsa */
 			chown_err(cxt->vcsan, uid, gid);
 		if (chmod(cxt->vcsan, cxt->tty_mode))
 			chmod_err(cxt->vcsan, cxt->tty_mode);
@@ -346,17 +348,16 @@ static void chown_tty(struct login_context *cxt)
 }
 
 /*
- * Reads the currect terminal path and initialize cxt->tty_* variables.
+ * Reads the current terminal path and initializes cxt->tty_* variables.
  */
 static void init_tty(struct login_context *cxt)
 {
-	const char *p;
 	struct stat st;
 	struct termios tt, ttt;
 
 	cxt->tty_mode = (mode_t) getlogindefs_num("TTYPERM", TTY_MODE);
 
-	cxt->tty_path = ttyname(0);		/* libc calls istty() here */
+	get_terminal_name(&cxt->tty_path, &cxt->tty_name, &cxt->tty_number);
 
 	/*
 	 * In case login is suid it was possible to use a hardlink as stdin
@@ -375,18 +376,6 @@ static void init_tty(struct login_context *cxt)
 		sleepexit(EXIT_FAILURE);
 	}
 
-	if (strncmp(cxt->tty_path, "/dev/", 5) == 0)
-		cxt->tty_name = cxt->tty_path + 5;
-	else
-		cxt->tty_name = cxt->tty_path;
-
-	for (p = cxt->tty_name; p && *p; p++) {
-		if (isdigit(*p)) {
-			cxt->tty_number = p;
-			break;
-		}
-	}
-
 #ifdef LOGIN_CHOWN_VCS
 	if (cxt->tty_number) {
 		/* find names of Virtual Console devices, for later mode change */
@@ -407,17 +396,17 @@ static void init_tty(struct login_context *cxt)
 	}
 
 	/* Kill processes left on this tty */
-	tcsetattr(0, TCSAFLUSH, &ttt);
+	tcsetattr(0, TCSANOW, &ttt);
 
 	/*
-	 * Let's close file decriptors before vhangup
+	 * Let's close file descriptors before vhangup
 	 * https://lkml.org/lkml/2012/6/5/145
 	 */
 	close(STDIN_FILENO);
 	close(STDOUT_FILENO);
 	close(STDERR_FILENO);
 
-	signal(SIGHUP, SIG_IGN);	/* so vhangup() wont kill us */
+	signal(SIGHUP, SIG_IGN);	/* so vhangup() won't kill us */
 	vhangup();
 	signal(SIGHUP, SIG_DFL);
 
@@ -430,14 +419,16 @@ static void init_tty(struct login_context *cxt)
 
 
 /*
- * Log failed login attempts in _PATH_BTMP if that exists.
+ * Logs failed login attempts in _PATH_BTMP, if it exists.
  * Must be called only with username the name of an actual user.
  * The most common login failure is to give password instead of username.
  */
 static void log_btmp(struct login_context *cxt)
 {
 	struct utmp ut;
+#if defined(_HAVE_UT_TV)        /* in <utmpbits.h> included by <utmp.h> */
 	struct timeval tv;
+#endif
 
 	memset(&ut, 0, sizeof(ut));
 
@@ -507,6 +498,7 @@ static void log_audit(struct login_context *cxt, int status)
 
 static void log_lastlog(struct login_context *cxt)
 {
+	struct sigaction sa, oldsa_xfsz;
 	struct lastlog ll;
 	time_t t;
 	int fd;
@@ -514,15 +506,20 @@ static void log_lastlog(struct login_context *cxt)
 	if (!cxt->pwd)
 		return;
 
+	/* lastlog is huge on systems with large UIDs, ignore SIGXFSZ */
+	memset(&sa, 0, sizeof(sa));
+	sa.sa_handler = SIG_IGN;
+	sigaction(SIGXFSZ, &sa, &oldsa_xfsz);
+
 	fd = open(_PATH_LASTLOG, O_RDWR, 0);
 	if (fd < 0)
-		return;
+		goto done;
 
 	if (lseek(fd, (off_t) cxt->pwd->pw_uid * sizeof(ll), SEEK_SET) == -1)
-		return;
+		goto done;
 
 	/*
-	 * Print last log message
+	 * Print last log message.
 	 */
 	if (!cxt->quiet) {
 		if (read(fd, (char *)&ll, sizeof(ll)) == sizeof(ll) &&
@@ -537,7 +534,8 @@ static void log_lastlog(struct login_context *cxt)
 				printf(_("on %.*s\n"),
 				       (int)sizeof(ll.ll_line), ll.ll_line);
 		}
-		lseek(fd, (off_t) cxt->pwd->pw_uid * sizeof(ll), SEEK_SET);
+		if (lseek(fd, (off_t) cxt->pwd->pw_uid * sizeof(ll), SEEK_SET) == -1)
+			goto done;
 	}
 
 	memset((char *)&ll, 0, sizeof(ll));
@@ -552,12 +550,15 @@ static void log_lastlog(struct login_context *cxt)
 
 	if (write_all(fd, (char *)&ll, sizeof(ll)))
 		warn(_("write lastlog failed"));
+done:
+	if (fd >= 0)
+		close(fd);
 
-	close(fd);
+	sigaction(SIGXFSZ, &oldsa_xfsz, NULL);		/* restore original setting */
 }
 
 /*
- * Update wtmp and utmp logs
+ * Update wtmp and utmp logs.
  */
 static void log_utmp(struct login_context *cxt)
 {
@@ -573,7 +574,7 @@ static void log_utmp(struct login_context *cxt)
 	 * login sometimes overwrites the runlevel entry in /var/run/utmp,
 	 * confusing sysvinit. I added a test for the entry type, and the
 	 * problem was gone. (In a runlevel entry, st_pid is not really a pid
-	 * but some number calculated from the previous and current runlevel).
+	 * but some number calculated from the previous and current runlevel.)
 	 * -- Michael Riepe <michael@stud.uni-hannover.de>
 	 */
 	while ((utp = getutent()))
@@ -583,8 +584,7 @@ static void log_utmp(struct login_context *cxt)
 			break;
 
 	/* If we can't find a pre-existing entry by pid, try by line.
-	 * BSD network daemons may rely on this.
-	 */
+	 * BSD network daemons may rely on this. */
 	if (utp == NULL && cxt->tty_name) {
 		setutent();
 		ut.ut_type = LOGIN_PROCESS;
@@ -593,7 +593,7 @@ static void log_utmp(struct login_context *cxt)
 	}
 
 	/* If we can't find a pre-existing entry by pid and line, try it by id.
-	 * Very stupid telnetd deamons don't set up utmp at all (kzak) */
+	 * Very stupid telnetd daemons don't set up utmp at all. (kzak) */
 	if (utp == NULL && cxt->tty_number) {
 	     setutent();
 	     ut.ut_type = DEAD_PROCESS;
@@ -673,22 +673,14 @@ static struct passwd *get_passwd_entry(const char *username,
 					 struct passwd *pwd)
 {
 	struct passwd *res = NULL;
-	size_t sz = 16384;
 	int x;
 
 	if (!pwdbuf || !username)
 		return NULL;
 
-#ifdef _SC_GETPW_R_SIZE_MAX
-	{
-		long xsz = sysconf(_SC_GETPW_R_SIZE_MAX);
-		if (xsz > 0)
-			sz = (size_t) xsz;
-	}
-#endif
-	*pwdbuf = xrealloc(*pwdbuf, sz);
+	*pwdbuf = xrealloc(*pwdbuf, UL_GETPW_BUFSIZ);
 
-	x = getpwnam_r(username, pwd, *pwdbuf, sz, &res);
+	x = getpwnam_r(username, pwd, *pwdbuf, UL_GETPW_BUFSIZ, &res);
 	if (!res) {
 		errno = x;
 		return NULL;
@@ -719,7 +711,7 @@ static void loginpam_err(pam_handle_t *pamh, int retcode)
 }
 
 /*
- * Composes "<host> login: " string; or returns "login: " is -H is given
+ * Composes "<host> login: " string; or returns "login: " if -H is given.
  */
 static const char *loginpam_get_prompt(struct login_context *cxt)
 {
@@ -745,7 +737,7 @@ static pam_handle_t *init_loginpam(struct login_context *cxt)
 
 	/*
 	 * username is initialized to NULL and if specified on the command line
-	 * it is set.  Therefore, we are safe not setting it to anything
+	 * it is set.  Therefore, we are safe not setting it to anything.
 	 */
 	rc = pam_start(cxt->remote ? "remote" : "login",
 		       cxt->username, &cxt->conv, &pamh);
@@ -757,8 +749,7 @@ static pam_handle_t *init_loginpam(struct login_context *cxt)
 	}
 
 	/* hostname & tty are either set to NULL or their correct values,
-	 * depending on how much we know
-	 */
+	 * depending on how much we know. */
 	rc = pam_set_item(pamh, PAM_RHOST, cxt->hostname);
 	if (is_pam_failure(rc))
 		loginpam_err(pamh, rc);
@@ -776,7 +767,7 @@ static pam_handle_t *init_loginpam(struct login_context *cxt)
 	if (is_pam_failure(rc))
 		loginpam_err(pamh, rc);
 
-	/* we need't the original username. We have to follow PAM. */
+	/* We don't need the original username. We have to follow PAM. */
 	free(cxt->username);
 	cxt->username = NULL;
 	cxt->pamh = pamh;
@@ -815,8 +806,8 @@ static void loginpam_auth(struct login_context *cxt)
 
 		if (rc == PAM_USER_UNKNOWN && !show_unknown)
 			/*
-			 * logging unknown usernames may be a security issue if
-			 * an user enter her password instead of her login name
+			 * Logging unknown usernames may be a security issue if
+			 * a user enters her password instead of her login name.
 			 */
 			cxt->username = NULL;
 		else
@@ -880,7 +871,7 @@ static void loginpam_acct(struct login_context *cxt)
 		loginpam_err(pamh, rc);
 
 	/*
-	 * Grab the user information out of the password file for future usage
+	 * Grab the user information out of the password file for future use.
 	 * First get the username that we are actually using, though.
 	 */
 	rc = loginpam_get_username(pamh, &cxt->username);
@@ -897,15 +888,15 @@ static void loginpam_acct(struct login_context *cxt)
 }
 
 /*
- * Note that position of the pam_setcred() call is discussable:
+ * Note that the position of the pam_setcred() call is discussable:
  *
- *  - the PAM docs recommends pam_setcred() before pam_open_session()
+ *  - the PAM docs recommend pam_setcred() before pam_open_session()
  *  - but the original RFC http://www.opengroup.org/rfc/mirror-rfc/rfc86.0.txt
  *    uses pam_setcred() after pam_open_session()
  *
  * The old login versions (before year 2011) followed the RFC. This is probably
- * not optimal, because there could be dependence between some session modules
- * and user's credentials.
+ * not optimal, because there could be a dependence between some session modules
+ * and the user's credentials.
  *
  * The best is probably to follow openssh and call pam_setcred() before and
  * after pam_open_session().                -- kzak@redhat.com (18-Nov-2011)
@@ -934,126 +925,8 @@ static void loginpam_session(struct login_context *cxt)
 }
 
 /*
- * We need to check effective UID/GID. For example $HOME could be on root
- * squashed NFS or on NFS with UID mapping and access(2) uses real UID/GID.
- * The open(2) seems as the surest solution.
- * -- kzak@redhat.com (10-Apr-2009)
- */
-static int effective_access(const char *path, int mode)
-{
-	int fd = open(path, mode);
-	if (fd != -1)
-		close(fd);
-	return fd == -1 ? -1 : 0;
-}
-
-/*
- * Check per accout or global hush-login setting.
- *
- * Hushed mode is enabled:
- *
- * a) if global (e.g. /etc/hushlogins) hush file exists:
- *     1) for ALL ACCOUNTS if the file is empty
- *     2) for the current user if the username or shell are found in the file
- *
- * b) if ~/.hushlogin file exists
- *
- * The ~/.hushlogin is ignored if the global hush file exists.
- *
- * The HUSHLOGIN_FILE login.def variable overwrites the default hush filename.
- *
- * Note that shadow-utils login(1) does not support "a1)". The "a1)" is
- * necessary if you want to use PAM for "Last login" message.
- *
- * -- Karel Zak <kzak@redhat.com> (26-Aug-2011)
- *
- *
- * Per-account check requires some explanation: As root we may not be able to
- * read the directory of the user if it is on an NFS mounted filesystem. We
- * temporarily set our effective uid to the user-uid making sure that we keep
- * root privs. in the real uid.
- *
- * A portable solution would require a fork(), but we rely on Linux having the
- * BSD setreuid()
- */
-static int get_hushlogin_status(struct passwd *pwd)
-{
-	const char *files[] = { _PATH_HUSHLOGINS, _PATH_HUSHLOGIN, NULL };
-	const char *file;
-	char buf[BUFSIZ];
-	int i;
-
-	file = getlogindefs_str("HUSHLOGIN_FILE", NULL);
-	if (file) {
-		if (!*file)
-			return 0;	/* empty HUSHLOGIN_FILE defined */
-
-		files[0] = file;
-		files[1] = NULL;
-	}
-
-	for (i = 0; files[i]; i++) {
-		int ok = 0;
-
-		file = files[i];
-
-		/* Global hush-file*/
-		if (*file == '/') {
-			struct stat st;
-			FILE *f;
-
-			if (stat(file, &st) != 0)
-				continue;	/* file does not exist */
-
-			if (st.st_size == 0)
-				return 1;	/* for all accounts */
-
-			f = fopen(file, "r");
-			if (!f)
-				continue;	/* ignore errors... */
-
-			while (ok == 0 && fgets(buf, sizeof(buf), f)) {
-				buf[strlen(buf) - 1] = '\0';
-				ok = !strcmp(buf, *buf == '/' ? pwd->pw_shell :
-								pwd->pw_name);
-			}
-			fclose(f);
-			if (ok)
-				return 1;	/* found username/shell */
-
-			return 0;		/* ignore per-account files */
-		}
-
-		/* Per-account setting */
-		if (strlen(pwd->pw_dir) + sizeof(file) + 2 > sizeof(buf))
-			continue;
-		else {
-			uid_t ruid = getuid();
-			gid_t egid = getegid();
-
-			sprintf(buf, "%s/%s", pwd->pw_dir, file);
-
-			if (setregid(-1, pwd->pw_gid) == 0 &&
-			    setreuid(0, pwd->pw_uid) == 0)
-				ok = effective_access(buf, O_RDONLY) == 0;
-
-			if (setuid(0) != 0 ||
-			    setreuid(ruid, 0) != 0 ||
-			    setregid(-1, egid) != 0) {
-				syslog(LOG_ALERT, _("hush login status: restore original IDs failed"));
-				exit(EXIT_FAILURE);
-			}
-			if (ok)
-				return 1;	/* enabled by user */
-		}
-	}
-
-	return 0;
-}
-
-/*
- * Detach the controlling terminal, fork, restore syslog stuff and create a new
- * session.
+ * Detach the controlling terminal, fork, restore syslog stuff, and create
+ * a new session.
  */
 static void fork_session(struct login_context *cxt)
 {
@@ -1070,15 +943,15 @@ static void fork_session(struct login_context *cxt)
 	sigaction(SIGHUP, &sa, &oldsa_hup);	/* ignore when TIOCNOTTY */
 
 	/*
-	 * detach the controlling tty
-	 * -- we needn't the tty in parent who waits for child only.
-	 *    The child calls setsid() that detach from the tty as well.
+	 * Detach the controlling tty.
+	 * We don't need the tty in a parent who only waits for a child.
+	 * The child calls setsid() that detaches from the tty as well.
 	 */
 	ioctl(0, TIOCNOTTY, NULL);
 
 	/*
-	 * We have care about SIGTERM, because leave PAM session without
-	 * pam_close_session() is pretty bad thing.
+	 * We have to beware of SIGTERM, because leaving a PAM session
+	 * without pam_close_session() is a pretty bad thing.
 	 */
 	sa.sa_handler = sig_handler;
 	sigaction(SIGHUP, &sa, NULL);
@@ -1087,14 +960,11 @@ static void fork_session(struct login_context *cxt)
 	closelog();
 
 	/*
-	 * We must fork before setuid() because we need to call
+	 * We must fork before setuid(), because we need to call
 	 * pam_close_session() as root.
 	 */
 	child_pid = fork();
 	if (child_pid < 0) {
-		/*
-		 * fork() error
-		 */
 		warn(_("fork failed"));
 
 		pam_setcred(cxt->pamh, PAM_DELETE_CRED);
@@ -1104,7 +974,7 @@ static void fork_session(struct login_context *cxt)
 
 	if (child_pid) {
 		/*
-		 * parent - wait for child to finish, then cleanup session
+		 * parent - wait for child to finish, then clean up session
 		 */
 		close(0);
 		close(1);
@@ -1157,39 +1027,44 @@ static void fork_session(struct login_context *cxt)
 static void init_environ(struct login_context *cxt)
 {
 	struct passwd *pwd = cxt->pwd;
-	char *termenv = NULL, **env;
+	char *termenv, **env;
 	char tmp[PATH_MAX];
 	int len, i;
 
 	termenv = getenv("TERM");
-	termenv = termenv ? xstrdup(termenv) : "dumb";
+	if (termenv)
+		termenv = xstrdup(termenv);
 
 	/* destroy environment unless user has requested preservation (-p) */
 	if (!cxt->keep_env) {
-		environ = (char **) xmalloc(sizeof(char *));
+		environ = xmalloc(sizeof(char *));
 		memset(environ, 0, sizeof(char *));
 	}
 
-	setenv("HOME", pwd->pw_dir, 0);	/* legal to override */
-	setenv("USER", pwd->pw_name, 1);
-	setenv("SHELL", pwd->pw_shell, 1);
-	setenv("TERM", termenv, 1);
+	xsetenv("HOME", pwd->pw_dir, 0);	/* legal to override */
+	xsetenv("USER", pwd->pw_name, 1);
+	xsetenv("SHELL", pwd->pw_shell, 1);
+	xsetenv("TERM", termenv ? termenv : "dumb", 1);
+	free(termenv);
 
-	if (pwd->pw_uid)
-		logindefs_setenv("PATH", "ENV_PATH", _PATH_DEFPATH);
+	if (pwd->pw_uid) {
+		if (logindefs_setenv("PATH", "ENV_PATH", _PATH_DEFPATH) != 0)
+			err(EXIT_FAILURE, _("failed to set the %s environment variable"), "PATH");
 
-	else if (logindefs_setenv("PATH", "ENV_ROOTPATH", NULL) != 0)
-		logindefs_setenv("PATH", "ENV_SUPATH", _PATH_DEFPATH_ROOT);
+	} else if (logindefs_setenv("PATH", "ENV_ROOTPATH", NULL) != 0 &&
+		   logindefs_setenv("PATH", "ENV_SUPATH", _PATH_DEFPATH_ROOT) != 0) {
+			err(EXIT_FAILURE, _("failed to set the %s environment variable"), "PATH");
+	}
 
 	/* mailx will give a funny error msg if you forget this one */
 	len = snprintf(tmp, sizeof(tmp), "%s/%s", _PATH_MAILDIR, pwd->pw_name);
-	if (len > 0 && (size_t) len + 1 <= sizeof(tmp))
-		setenv("MAIL", tmp, 0);
+	if (len > 0 && (size_t) len < sizeof(tmp))
+		xsetenv("MAIL", tmp, 0);
 
 	/* LOGNAME is not documented in login(1) but HP-UX 6.5 does it. We'll
 	 * not allow modifying it.
 	 */
-	setenv("LOGNAME", pwd->pw_name, 1);
+	xsetenv("LOGNAME", pwd->pw_name, 1);
 
 	env = pam_getenvlist(cxt->pamh);
 	for (i = 0; env && env[i]; i++)
@@ -1197,7 +1072,7 @@ static void init_environ(struct login_context *cxt)
 }
 
 /*
- * Called for -h option, initialize cxt->{hostname,hostaddress}
+ * This is called for the -h option, initializes cxt->{hostname,hostaddress}.
  */
 static void init_remote_info(struct login_context *cxt, char *remotehost)
 {
@@ -1244,20 +1119,28 @@ int main(int argc, char **argv)
 	char *buff;
 	int childArgc = 0;
 	int retcode;
+	struct sigaction act;
 
 	char *pwdbuf = NULL;
 	struct passwd *pwd = NULL, _pwd;
 
 	struct login_context cxt = {
-		.tty_mode = TTY_MODE,		/* tty chmod() */
-		.pid = getpid(),		/* PID */
-		.conv = { misc_conv, NULL }	/* PAM conversation function */
+		.tty_mode = TTY_MODE,		  /* tty chmod() */
+		.pid = getpid(),		  /* PID */
+#ifdef HAVE_SECURITY_PAM_MISC_H
+		.conv = { misc_conv, NULL }	  /* Linux-PAM conversation function */
+#elif defined(HAVE_SECURITY_OPENPAM_H)
+		.conv = { openpam_ttyconv, NULL } /* OpenPAM conversation function */
+#endif
+
 	};
 
 	timeout = (unsigned int)getlogindefs_num("LOGIN_TIMEOUT", LOGIN_TIMEOUT);
 
 	signal(SIGALRM, timedout);
-	siginterrupt(SIGALRM, 1);	/* we have to interrupt syscalls like ioclt() */
+	(void) sigaction(SIGALRM, NULL, &act);
+	act.sa_flags &= ~SA_RESTART;
+	sigaction(SIGALRM, &act, NULL);
 	alarm(timeout);
 	signal(SIGQUIT, SIG_IGN);
 	signal(SIGINT, SIG_IGN);
@@ -1288,7 +1171,7 @@ int main(int argc, char **argv)
 		case 'h':
 			if (getuid()) {
 				fprintf(stderr,
-					_("login: -h for super-user only.\n"));
+					_("login: -h is for superuser only\n"));
 				exit(EXIT_FAILURE);
 			}
 			init_remote_info(&cxt, optarg);
@@ -1303,7 +1186,9 @@ int main(int argc, char **argv)
 			return EXIT_SUCCESS;
 		case '?':
 		default:
-			fprintf(stderr, _("usage: login [ -p ] [ -h host ] [ -H ] [ -f username | username ]\n"));
+			fprintf(stderr, _("Usage: login [-p] [-h <host>] [-H] [[-f] <username>]\n"));
+			fputs(USAGE_SEPARATOR, stderr);
+			fputs(_("Begin a session on the system.\n"), stderr);
 			exit(EXIT_FAILURE);
 		}
 	argc -= optind;
@@ -1313,8 +1198,8 @@ int main(int argc, char **argv)
 		char *p = *argv;
 		cxt.username = xstrdup(p);
 
-		/* wipe name - some people mistype their password here */
-		/* (of course we are too late, but perhaps this helps a little ..) */
+		/* Wipe the name - some people mistype their password here. */
+		/* (Of course we are too late, but perhaps this helps a little...) */
 		while (*p)
 			*p++ = ' ';
 	}
@@ -1323,10 +1208,10 @@ int main(int argc, char **argv)
 		close(cnt);
 
 	setpgrp();	 /* set pgid to pid this means that setsid() will fail */
+	init_tty(&cxt);
 
 	openlog("login", LOG_ODELAY, LOG_AUTHPRIV);
 
-	init_tty(&cxt);
 	init_loginpam(&cxt);
 
 	/* login -f, then the user has already been authenticated */
@@ -1338,8 +1223,8 @@ int main(int argc, char **argv)
 	/*
 	 * Authentication may be skipped (for example, during krlogin, rlogin,
 	 * etc...), but it doesn't mean that we can skip other account checks.
-	 * The account could be disabled or password expired (although
-	 * kerberos ticket is valid).         -- kzak@redhat.com (22-Feb-2006)
+	 * The account could be disabled or the password has expired (although
+	 * the kerberos ticket is valid).      -- kzak@redhat.com (22-Feb-2006)
 	 */
 	loginpam_acct(&cxt);
 
@@ -1356,10 +1241,9 @@ int main(int argc, char **argv)
 
 	/*
 	 * Initialize the supplementary group list. This should be done before
-	 * pam_setcred because the PAM modules might add groups during
-	 * pam_setcred.
+	 * pam_setcred, because PAM modules might add groups during that call.
 	 *
-         * For root we don't call initgroups, instead we call setgroups with
+	 * For root we don't call initgroups, instead we call setgroups with
 	 * group 0. This avoids the need to step through the whole group file,
 	 * which can cause problems if NIS, NIS+, LDAP or something similar
 	 * is used and the machine has network problems.
@@ -1374,7 +1258,7 @@ int main(int argc, char **argv)
 	}
 
 	/*
-	 * Open PAM session (after successful authentication and account check)
+	 * Open PAM session (after successful authentication and account check).
 	 */
 	loginpam_session(&cxt);
 
@@ -1383,7 +1267,7 @@ int main(int argc, char **argv)
 
 	endpwent();
 
-	cxt.quiet = get_hushlogin_status(pwd);
+	cxt.quiet = get_hushlogin_status(pwd, 1);
 
 	log_utmp(&cxt);
 	log_audit(&cxt, 1);
@@ -1431,12 +1315,12 @@ int main(int argc, char **argv)
 	}
 
 	/*
-	 * Detach the controlling terminal, fork() and create, new session
-	 * and reinilizalize syslog stuff.
+	 * Detach the controlling terminal, fork, and create a new session
+	 * and reinitialize syslog stuff.
 	 */
 	fork_session(&cxt);
 
-	/* discard permissions last so can't get killed and drop core */
+	/* discard permissions last so we can't get killed and drop core */
 	if (setuid(pwd->pw_uid) < 0 && pwd->pw_uid) {
 		syslog(LOG_ALERT, _("setuid() failed"));
 		exit(EXIT_FAILURE);
@@ -1486,5 +1370,3 @@ int main(int argc, char **argv)
 
 	exit(EXIT_SUCCESS);
 }
-
-
