@@ -31,13 +31,22 @@
 
 /* Sccsid @(#)pg.c 1.44 (gritter) 2/8/02 - modified for util-linux */
 
+/*
+ * This command is deprecated.  The utility is in maintenance mode,
+ * meaning we keep them in source tree for backward compatibility
+ * only.  Do not waste time making this command better, unless the
+ * fix is about security or other very critical issue.
+ *
+ * See Documentation/deprecated.txt for more information.
+ */
+
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
 #ifndef	TIOCGWINSZ
 # include <sys/ioctl.h>
 #endif
-#include <sys/termios.h>
+#include <termios.h>
 #include <fcntl.h>
 #include <regex.h>
 #include <stdio.h>
@@ -49,7 +58,6 @@
 #include <unistd.h>
 #include <signal.h>
 #include <setjmp.h>
-#include <libgen.h>
 
 #ifdef HAVE_NCURSES_H
 # include <ncurses.h>
@@ -68,7 +76,7 @@
 
 #define	READBUF		LINE_MAX	/* size of input buffer */
 #define CMDBUF		255		/* size of command buffer */
-#define	TABSIZE		8		/* spaces consumed by tab character */
+#define	PG_TABSIZE	8		/* spaces consumed by tab character */
 
 #define	cuc(c)		((c) & 0377)
 
@@ -213,6 +221,10 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	fprintf(out,
 		_(" %s [options] [+line] [+/pattern/] [files]\n"),
 		program_invocation_short_name);
+
+	fputs(USAGE_SEPARATOR, out);
+	fputs(_("Browse pagewise through text files.\n"), out);
+
 	fputs(USAGE_OPTIONS, out);
 	fputs(_(" -number      lines per page\n"), out);
 	fputs(_(" -c           clear screen before displaying\n"), out);
@@ -224,8 +236,11 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	fputs(_(" -s           print messages to stdout\n"), out);
 	fputs(_(" +number      start at the given line\n"), out);
 	fputs(_(" +/pattern/   start at the line containing pattern\n"), out);
-	fputs(_(" -h           display this help and exit\n"), out);
-	fputs(_(" -V           output version information and exit\n"), out);
+
+	fputs(USAGE_SEPARATOR, out);
+	fputs(USAGE_HELP, out);
+	fputs(USAGE_VERSION, out);
+
 	fprintf(out, USAGE_MAN_TAIL("pg(1)"));
 	quit(out == stderr ? 2 : 0);
 }
@@ -402,7 +417,7 @@ static char *endline_for_mb(unsigned col, char *s)
 			goto ended;
 			/* Cursor right. */
 		case L'\t':
-			pos += TABSIZE - (pos % TABSIZE);
+			pos += PG_TABSIZE - (pos % PG_TABSIZE);
 			break;
 		default:
 			if (iswprint(*p))
@@ -466,7 +481,7 @@ static char *endline(unsigned col, char *s)
 			goto cend;
 			/* Cursor right. */
 		case '\t':
-			pos += TABSIZE - (pos % TABSIZE);
+			pos += PG_TABSIZE - (pos % PG_TABSIZE);
 			break;
 		default:
 			pos++;
@@ -790,7 +805,7 @@ static char *colb(char *s)
 }
 
 #ifdef HAVE_WIDECHAR
-/* Convert nonprintable characters to spaces in case MB_CUR_MAX > 1.  */
+/* Convert non-printable characters to spaces in case MB_CUR_MAX > 1.  */
 static void makeprint_for_mb(char *s, size_t l)
 {
 	char *t = s;
@@ -810,7 +825,7 @@ static void makeprint_for_mb(char *s, size_t l)
 }
 #endif
 
-/* Convert nonprintable characters to spaces. */
+/* Convert non-printable characters to spaces. */
 static void makeprint(char *s, size_t l)
 {
 #ifdef HAVE_WIDECHAR
@@ -923,7 +938,7 @@ static void pgfile(FILE *f, const char *name)
 	}
 	find = tmpfile();
 	if (fbuf == NULL || find == NULL) {
-		warn(_("Cannot create tempfile"));
+		warn(_("Cannot create temporary file"));
 		quit(++exitstatus);
 	}
 	if (searchfor) {
@@ -1228,7 +1243,14 @@ static void pgfile(FILE *f, const char *name)
 					/* No error check for compat. */
 					fwrite_all(b, sizeof *b, sz, save);
 				}
-				fclose(save);
+				if (close_stream(save) != 0) {
+					cmd.count = errno;
+					mesg(_("write failed"));
+					mesg(": ");
+					mesg(p);
+					mesg(strerror(cmd.count));
+					goto newcmd;
+				}
 				fseeko(fbuf, (off_t)0, SEEK_END);
 				mesg(_("saved"));
 				goto newcmd;
@@ -1512,7 +1534,7 @@ int main(int argc, char **argv)
 		  _("%s %s Copyright (c) 2000-2001 Gunnar Ritter. All rights reserved.\n"),
 		  program_invocation_short_name, PACKAGE_VERSION);
 
-	setlocale(LC_MESSAGES, "");
+	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 	atexit(close_stdout);
@@ -1535,6 +1557,16 @@ int main(int argc, char **argv)
 		if (*argv[arg] != '-' || argv[arg][1] == '\0')
 			break;
 		argc--;
+
+		if (!strcmp(argv[arg], "--help")) {
+		    usage(stdout);
+		}
+
+		if (!strcmp(argv[arg], "--version")) {
+		    printf(UTIL_LINUX_VERSION);
+		    return EXIT_SUCCESS;
+		}
+
 		for (i = 1; argv[arg][i]; i++) {
 			switch (argv[arg][i]) {
 			case '-':
